#!/bin/sh

# Copyright 2017 - 2018 Paul Hänsch
#
# This is CGIlite.
# A collection of posix shell functions for writing CGI scripts.
# 
# CGIlite is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# CGIlite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with CGIlite.  If not, see <http://www.gnu.org/licenses/>. 

# ksh and zsh workaround
# set -o posix # ksh, not portable
setopt -o OCTAL_ZEROES 2>&-

CR="
"
BR='
'
cgilite_timeout=2

HEADER(){
  # Read value of header line. Use this instead of
  # referencing HTTP_* environment variables.
  if [ -n "${cgilite_headers+x}" ]; then
    printf %s "$cgilite_headers" \
    | sed -rn 's;^'"${1}"': ([^\r]+)\r?$;\1;i; tX; d; :X;p;q;'
  else
    eval "printf %s \"\$HTTP_$(printf %s "${1}" |tr a-z A-Z |tr -c A-Z _)\""
  fi
}

HEX_DECODE='
  s;\\;\\\\;g; :HEXDECODE_X; s;%([^0-9A-F]);\\045\1;g; tHEXDECODE_X;
  # Hexadecimal { %00 - %FF } will be transformed to octal { \000 - \377 } for posix printf
  s;%[0123].;&\\0;g; s;%[4567].;&\\1;g; s;%[89AB].;&\\2;g; s;%[CDEF].;&\\3;g;
  s;%[048C][0-7]\\.;&0;g; s;%[048C][89A-F]\\.;&1;g; s;%[159D][0-7]\\.;&2;g; s;%[159D][89A-F]\\.;&3;g;
  s;%[26AE][0-7]\\.;&4;g; s;%[26AE][89A-F]\\.;&5;g; s;%[37BF][0-7]\\.;&6;g; s;%[37BF][89A-F]\\.;&7;g;
  s;%.[08](\\..);\10;g; s;%.[19](\\..);\11;g; s;%.[2A](\\..);\12;g; s;%.[3B](\\..);\13;g;
  s;%.[4C](\\..);\14;g; s;%.[5D](\\..);\15;g; s;%.[6E](\\..);\16;g; s;%.[7F](\\..);\17;g;
'

HEX_DECODE(){
  printf "$(printf %s "$1" |sed -r "$HEX_DECODE")"
}

if [ -z "$REQUEST_METHOD" ]; then
  # no webserver variables means we are running via inetd / ncat
  # so use builtin web server

  # Use env from inetd as webserver variables
  REMOTE_ADDR="${TCPREMOTEIP}"
  SERVER_NAME="${TCPLOCALIP}"
  SERVER_PORT="${TCPLOCALPORT}"

  # Wait 2 seconds for request or kill connection through watchdog.
  # Once Request is received the watchdog will be suspended (killed).
  # At the end of the loop the watchdog will be restarted to enable
  # timeout for the subsequent request.

  (sleep $cgilite_timeout && kill $$) & cgilite_watchdog=$!
  while read REQUEST_METHOD REQUEST_URI SERVER_PROTOCOL; do
    kill $cgilite_watchdog
    PATH_INFO="$(HEX_DECODE "${REQUEST_URI%\?*}")"
    QUERY_STRING="${REQUEST_URI#*\?}"
    cgilite_headers="$(while read -r hl; do [ "${hl%${CR}}" ] && printf '%s\n' "$hl" || break; done )"

    HTTP_CONTENT_LENGTH="$(HEADER Content-Length |grep -xE '[0-9]+')"

    export REMOTE_ADDR SERVER_NAME SERVER_PORT REQUEST_METHOD REQUEST_URI SERVER_PROTOCOL \
           PATH_INFO QUERY_STRING HTTP_CONTENT_LENGTH

    # Try to serve multiple requests, provided that script serves a
    # Content-Length header.
    # Without Content-Length header, connection will terminate after
    # script.

    cgilite_status='200 OK'; cgilite_response=''; cgilite_cl="Connection: close${CR}";
    . "$0" | while read -r l; do case $l in
      Status:*) cgilite_status="${l#Status: }";;
      Content-Length:*) cgilite_cl="${l}";;
      $CR) printf '%s %s\r\n%s\n%s\n\r\n' \
             'HTTP/1.1' "${cgilite_status%${CR}}" \
             "$cgilite_response" "${cgilite_cl}"
           cat
           [ "${cgilite_cl#Connection}" = "${cgilite_cl}" ]; exit;;
      *) cgilite_response="$cgilite_response${BR}$l";;
    esac; done || exit 0;
    (sleep $cgilite_timeout && kill $$) & cgilite_watchdog=$!
  done
fi

if [ "$REQUEST_METHOD" = POST -a "${HTTP_CONTENT_LENGTH:=${CONTENT_LENGTH:=0}}" -gt 0 ]; then
  cgilite_post="$(head -c "$HTTP_CONTENT_LENGTH")"
fi

[ -n "${DEBUG+x}" ] && env

cgilite_count(){
  printf %s "&$1" \
  | grep -oE '&'"$2"'=[^&]*' \
  | wc -l
}

cgilite_value(){
  local str="&$1" name="$2" cnt="${3:-1}"
  while [ $cnt -gt 0 ]; do
    str=${str#*&${name}=}
    cnt=$((cnt - 1))
  done
  printf "$(printf %s "${str%%&*}" |sed -r 's;\+; ;g;'"$HEX_DECODE")"
}

GET(){ cgilite_value "${QUERY_STRING}" $@; }
GET_COUNT(){ cgilite_count "${QUERY_STRING}" $1; }

POST(){ cgilite_value "${cgilite_post}" $@; }
POST_COUNT(){ cgilite_count "${cgilite_post}" $1; }

REF(){ cgilite_value "${HTTP_REFERER#*\?}" $@; }
REF_COUNT(){ cgilite_count "${HTTP_REFERER#*\?}" $1; }

COOKIE(){
  HEX_DECODE "$(
    HEADER Cookie \
    | grep -oE '(^|; ?)'"$1"'=[^;]*' \
    | sed -rn "${2:-1}"'{s;^[^=]+=;;; s;\+; ;g; p;}'
  )"
}

HTML(){
  # HTML Entity Coding
  # Prints UTF-8 string as decimal Unicode Code Points
  # Useful for escaping user input for use in HTML text and attributes
  { [ $# -eq 0 ] && cat || printf %s "$*"; } \
  | hexdump -ve '/1 "%03o\n"' \
  | while read n; do
    case $n in
      # bitbanging octal UTF-8 chains into singular 7 digit octal numbers
      [01]??) printf '0000%s' $n;; # 7 bit ASCII character, nothing to do
      2??)    printf '%s' ${n#2};; # tail fragment, append 6 bit
      3[0123]?) printf '000%s' ${n#3};; # 2 octet (11 bit) chain start
      34?) printf '00%s' ${n#34};; # 3 octet (16 bit) chain start
      35?) printf '01%s' ${n#35};; # 3 octet chain start, high
      36?) printf '%s' ${n#36};;   # 4 octet (21 bit) chain start
    esac
  done \
  | sed -r 's;.{7};&\n;g;' \
  | while read n; do
    printf '&#%d;' $((0$n))
  done
}

URL(){
  # Code every character in URL escape hex format
  # except alphanumeric ascii

  { [ $# -eq 0 ] && cat || printf %s "$*"; } \
  | hexdump -v -e '/1 ",%02X"' \
  | sed 's;,;%;g; s;%2F;/;g;'
}

PATH(){
  { [ $# -eq 0 ] && cat || printf %s "$*"; } \
  | sed -r 's;^.*$;/&/;; s;/+;/;g;
            :X;
            s;^/\.\./;/;; s;/\./;/;g;
            tX;
            s;/[^/]+/\.\./;/;;
            tX;
            s;^(/.*)/$;\1;'
}


SET_COOKIE(){
  local expire cookie
  case "$1" in
    ''|0|session) expire='';;
    [+-][0-9]*)   expire="$(date -R -d @$(($(date +%s) + $1)))";;
    *)            expire="$(date -R -d "$1")";;
  esac
  cookie="$2"

  printf 'Set-Cookie: %s' "$cookie"
  [ -n "$expire" ] && printf '; Expires=%s' "${expire%+????}${expire:+GMT}"
  [ $# -ge 3 ] && shift 2 && printf '; %s' "$@"
  printf '\r\n'
}

REDIRECT(){
  printf '%s: %s\r\n' \
    Status "303 See Other" \
    Content-Length 0 \
    Location "$*"
  printf '\r\n'
  exit 0
}
